"""
This module contains the class PDF. It is used to generate a PDF report from a list of entries.
"""
import math
import urllib

import pandas as pd
from fpdf import FPDF, TitleStyle

from confectionary.report_generation import load_word2vec_model, split_to_pars
from confectionary.utils import find_text_keywords, fix_punct_spaces, get_timestamp


class PDF(FPDF):
    """
    PDF - instance of the class FPDF
    note, A4 dimensions in mm are: 210 wide x 297 long
    helpful doc: https://pyfpdf.github.io/fpdf2/ReferenceManual.html
    Parameters
    ----------
    FPDF : class, optional, default FPDF, the class FPDF
    """

    def __init__(
        self,
        orientation="P",
        unit="mm",
        format="A4",
        font_cache_dir=None,
        key_phrase="text2pdf",
        is_ewriter=False,
        split_paragraphs=True,
        bottom_margin=10,
    ):
        super().__init__(orientation, unit, format, font_cache_dir)
        self.key_phrase = key_phrase
        self.is_ewriter = is_ewriter
        self.split_paragraphs = split_paragraphs
        self.bottom_margin = bottom_margin
        self.initialized_word2vec = False
        self.wrdvecs = None

    def header(
        self, top_margin=10, header_text="", header_font_size=8, header_font_style="I"
    ):
        """
        header - add a header to the PDF

        Parameters
        ----------
        top_margin : int, optional, default 10, the top margin of the header
        header_text : str, optional, default "", the text to be displayed in the header
        header_font_size : int, optional, default 8, the font size of the header text
        header_font_style : str, optional, default "I", the font style of the header text
        """
        self.set_y(top_margin)  # set 1 CM from the top
        self.set_font("helvetica", header_font_style, header_font_size)
        th = self.font_size  # Text height

        self.cell(
            0,
            h=th,
            txt=f"{header_text} Created {get_timestamp()}",
            border="",
            ln=1,
            align="R",
            fill=False,
        )

        # add a rectangular border based on margins
        x_margin_L = self.l_margin
        x_margin_R = self.r_margin
        y_margin_T = self.t_margin
        y_margin_B = self.b_margin
        footer_location = th + self.bottom_margin  # where to end on bottom of page
        usable_x = self.w - x_margin_L - x_margin_R
        usable_y = self.h - footer_location - y_margin_T
        start_box = top_margin + th
        self.rect(x=x_margin_L, y=start_box, w=usable_x, h=usable_y, style="D")

    def footer(
        self, footer_text="", footer_font_size=8, footer_font_style="I", text_color=128
    ):
        """
        footer - add a footer to the PDF

        Parameters
        ----------
        footer_text : str, optional, default "", the text to be displayed in the footer
        footer_font_size : int, optional, default 8, the font size of the footer text
        footer_font_style : str, optional, default "I", the font style of the footer text
        text_color : int, optional, default 128, the color of the footer text
        """
        # Position at 1 cm from bottom
        self.set_y(-self.bottom_margin)  # Position at 1 cm from bottom
        # helvetica italic 8
        self.set_font("helvetica", footer_font_style, footer_font_size)
        # Text color in gray
        self.set_text_color(text_color)
        # Text height
        th = self.font_size

        # set link to TOC
        TOC_link = self.add_link()
        self.set_link(TOC_link, page=1)

        total_footer = f"{footer_text} - {self.key_phrase} - Page {self.page_no()}"
        self.cell(w=0, h=th, txt=total_footer, border=0, ln=0, link=TOC_link, align="R")

    def update_margins(self):
        """
        update_margins - update the margins of the PDF, based assigned margins
        """
        if self.is_ewriter:
            print("EWriter mode - setting margins large")

            # allocate margins
            pdf_width = self.w
            divided_space = math.floor(pdf_width / 3)
            divided_space = divided_space * 0.75  # take a lower %

            leftright_width = divided_space

            self.l_margin = leftright_width
            self.r_margin = leftright_width

            self.t_margin = 20
            self.b_margin = 20
        else:
            self.l_margin = 10
            self.r_margin = 10

            self.t_margin = 20
            self.b_margin = 20

    def update_title_formats(
        self,
        font_family="Helvetica",
        main_title_font_size=24,
        main_title_font_style="B",
        sub_title_font_size=20,
        sub_title_font_style="B",
        do_underline=False,
    ):
        """
        update_title_formats - update the title formats of the PDF, based assigned formats

        Parameters
        ----------
        font_family : str, optional, default "Helvetica", the font family of the title
        main_title_font_size : int, optional, default 24, the font size of the main title
        main_title_font_style : str, optional, default "B", the font style of the main title
        sub_title_font_size : int, optional, default 20, the font size of the sub title
        sub_title_font_style : str, optional, default "B", the font style of the sub title
        do_underline : bool, optional, default False, whether to underline the title
        """

        self.set_section_title_styles(
            # Level 0 titles:
            TitleStyle(
                font_family=font_family,
                font_style=main_title_font_style,
                font_size_pt=main_title_font_size,
                color=0,  # black
                underline=do_underline,
                t_margin=10,
                l_margin=10,
                b_margin=0,
            ),
            # Level 1 subtitles:
            TitleStyle(
                font_family=font_family,
                font_style=sub_title_font_style,
                font_size_pt=sub_title_font_size,
                color=128,
                underline=do_underline,
                t_margin=10,
                l_margin=20,
                b_margin=5,
            ),
        )

    def init_word2vec(self):
        """
        init_word2vec - initialize the word2vec model
        """
        if self.split_paragraphs and not self.initialized_word2vec:
            gensim_model = load_word2vec_model()
            self.wrdvecs = pd.DataFrame(
                gensim_model.vectors, index=gensim_model.key_to_index
            )
            self.initialized_word2vec = True

    def chapter_title(self, num: int, label: str):
        """
        chapter_title - add a chapter title to the PDF

        Parameters
        ----------
        num : int, the chapter number
        label : str, the chapter label
        """
        self.set_font("Helvetica", "B", 14)
        self.set_fill_color(200, 220, 255)
        th = self.font_size
        total_title = f"Chapter {num} - {label.replace('_', ' ').strip()}"
        self.start_section(total_title)
        self.ln(4)

    def chapter_body_filepath(self, filepath, verbose=False, std_font=14):
        """
        chapter_body_filepath - read in a file and add to the PDF as a chapter

        Parameters
        ----------
        filepath : str, filepath to the file to be added to the PDF
        verbose : bool, optional, by default False
        std_font : int, optional, by default 14
        """

        with open(filepath, "r", encoding="UTF-8", errors="ignore") as f:
            text = f.read()  # Read text file
        text = fix_punct_spaces(text)
        # Times for reading
        self.set_font("Times", "", std_font)
        # Text height. Add some millimeters for legibility
        th = self.font_size + 0.75

        enc_line = text.encode("latin-1", errors="replace")
        decoded_line = enc_line.decode("latin-1")
        if self.split_paragraphs:
            self.init_word2vec()
            paragraph_list = split_to_pars(
                decoded_line, wordvectors=self.wrdvecs, use_punkt=True
            )
            for par in paragraph_list:
                formatted_par = " " * 8 + par.strip()
                self.multi_cell(w=0, h=th, txt=formatted_par, border=0)
                self.ln()
        else:
            # "keep all text as one blob"

            self.multi_cell(w=0, h=th, txt=decoded_line, border=0)
            self.ln()

        if len(text) > 1500:
            # add keywords generated by YAKE
            self.generic_text(f"KeyWords:\n\t{find_text_keywords(text)}")
            self.ln()

        self.set_font("", "I")
        self.cell(0, th, "(end of excerpt)")

    def chapter_body_fromURL(self, aURL):
        """
        chapter_body_fromURL - read in a URL and add to the PDF as a chapter

        Parameters
        ----------
        aURL : str, the URL to be added to the PDF
        """

        session = urllib.request.urlopen(aURL)
        self.set_font("Times", "", 10)
        # Text height
        th = self.font_size

        for line in session:
            URL_line = line.decode("utf-8")
            enc_line = URL_line.encode("latin-1", errors="replace")
            decoded_line = enc_line.decode("latin-1")
        if self.split_paragraphs:
            self.init_word2vec()
            paragraph_list = split_to_pars(
                decoded_line, wordvectors=self.wrdvecs, use_punkt=True
            )
            for par in paragraph_list:
                formatted_par = " " * 8 + par
                self.multi_cell(w=0, h=th, txt=formatted_par, border=0)
                self.ln()
        else:
            # "keep all text as one blob"

            self.multi_cell(w=0, h=th, txt=decoded_line, border=0)
            self.ln()

        # Mention in italics
        self.set_font("", "I")
        self.cell(0, 5, "(end of excerpt)")

    def print_chapter(self, filepath, num: int, title: str = ""):
        """
        print_chapter is a wrapper for chapter_body that takes a filepath, title, and number, and prints the chapter

        :param fname: filepath to text file
        :param num: chapter number
        :param title: chapter title
        """
        self.add_page()
        self.chapter_title(num, title)
        self.chapter_body_filepath(filepath)

    def print_chapter_URL(self, theURL, num: int, title: str = ""):
        """
        print_chapter_URL is a wrapper for chapter_body that takes a URL, title, and number, and prints the chapter from the URL

        Parameters
        ----------
        theURL : str, URL
        num : int, chapter number, starting at 1
        title : str, optional, title of chapter
        """
        self.add_page()
        self.chapter_title(num, title)
        self.chapter_body_fromURL(theURL)

    def figure_title(self, title: str):
        """
        figure_title - add a figure title to the PDF
        """
        self.set_font("helvetica", "B", 14)
        th = self.font_size  # Text height
        self.multi_cell(w=0, h=th, txt=title, border="B", ln=1, align="L", fill=False)
        self.ln()  # Line break

    def write_big_title(self, a_title: str, font_size=24, font_style="B"):
        """
        write_big_title - write a big title to the PDF, typically the title of the document.

        Parameters
        ----------
        a_title : str, the title to be written
        font_size : int, optional, by default 24, the font size to be used
        font_style : str, optional, by default "B", the font style to be used
        """

        self.set_font("Helvetica", font_style, font_size)
        th = self.font_size  # Text height
        self.multi_cell(w=0, h=th, txt=a_title, border="B", ln=1, align="C", fill=False)
        self.ln()

    def generic_text(self, the_text: str, the_size: int = 12):
        """
        generic_text is a wrapper for multi_cell that takes a string and font size, and prints it in the current page

        Parameters
        ----------
        the_text : str, required, the text to be printed
        the_size : int, optional, the font size to be used, default is 12
        """
        self.set_font("Times", "", int(the_size))

        # Text height
        th = self.font_size
        self.multi_cell(w=0, h=th, txt=the_text, ln=1, align="L")

    def comment_text(
        self, the_text: str, the_size: int = 12, preamble="**PLEASE NOTE:** "
    ):
        """
        comment_text is a wrapper for generic_text that adds a comment to the pdf.

        Parameters
        ----------
        the_text : str, the text to be added
        the_size : int, optional, the font size of the text. The default is 12.
        preamble : str, optional, the text to be added before the comment. The default is "**PLEASE NOTE:** ".
        """
        self.set_font("Courier", size=int(the_size), style="B")
        self.set_text_color(r=153, g=76, b=0)
        # Text height
        th = self.font_size
        my_text = preamble + the_text
        enc_line = my_text.encode("latin-1", errors="replace")
        decoded_line = enc_line.decode("latin-1")
        self.multi_cell(w=0, h=th, txt=decoded_line, ln=1, align="L")
        self.set_text_color(r=0, g=0, b=0)



import glob
import os
import setuptools
from pathlib import Path


def get_package_description():
    """Returns a description of this package from the markdown files."""
    _readme = Path("README.md")
    _history = Path("HISTORY.md")
    if _readme.exists() and _history.exists():
        with open(_readme.resolve(), "r", encoding="utf-8", errors="ignore") as f:
            readme = f.read()
    else:
        readme = "README"
    if _history.exists():
        with open(_history.resolve(), "r", encoding="utf-8", errors="ignore") as f:
            history = f.read()
    else:
        history = "No history yet."
    return f"{readme}\n\n{history}"


def get_scripts_from_bin():
    """Get all local scripts from bin so they are included in the package."""
    return glob.glob("bin/*")


def get_requirements():
    """Returns all requirements for this package."""
    with open("requirements.txt", "r", encoding="utf-8") as f:
        requirements = f.readlines()
    return list(requirements)

def scour_for_file(file_name:str):
    """
    scour_for_file - search every possible location for a file name. Load each line from that filename into a list.
    """
    contents = []
    for root, dirs, files in os.walk("."):
        if file_name in files:
            with open(os.path.join(root, file_name), "r") as f:
                contents = f.readlines()
    assert len(contents) > 0, f"Could not find {file_name} in any of the locations"
    file_contents = [l for l in contents if l.strip()]
    return file_contents




try:
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()
except FileNotFoundError as e:
    print(f"could not read README.md: {e}")
    long_description = get_package_description()


# requirements = get_requirements()
# scripts = get_scripts_from_bin()


setuptools.setup(
    name="confectionary",
    author="Peter Szemraj, Jonathan Lehner",
    author_email="szemraj.dev@gmail.com",
    description="A tool to quickly create sweet PDF files from text files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pszemraj/confectionary",
    include_package_data=True,
    include_dirs=["bin"],
    # package_dir={"": "confectionary"},
    data_files=[("", ["LICENSE"]), ("", ["requirements.txt"]), ("", ["README.md"])],
    packages=setuptools.find_packages(),
    install_requires=scour_for_file("requirements.txt"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Topic :: Text Processing",
    ],
    scripts=get_scripts_from_bin(),
    python_requires=">=3.7",
    setuptools_git_versioning={
        "enabled": True,
    },
    setup_requires=["setuptools-git-versioning"],
)



"""
this script contains helper functions to generate a PDF report from a list of entries, including paragraphs splitting by coherence, and a table of contents.
"""
import math
import pickle
import warnings
from pathlib import Path

import gensim.downloader as api
import nltk
from sklearn.feature_extraction.text import CountVectorizer
from textsplit.algorithm import split_optimal
from textsplit.tools import SimpleSentenceTokenizer, get_penalty, get_segments
from tqdm.auto import tqdm

model_storage_loc = Path.cwd() / "models"
model_storage_loc.mkdir(exist_ok=True)


def load_word2vec_model(
    word2vec_mname="word2vec-google-news-300",
    storage_loc=model_storage_loc,
    verbose=False,
):
    """
    Loads a word2vec model from the Google News corpus.
    :param model_name: Name of the model to load.
    :return: Word2Vec model.
    """
    storage_loc = Path(storage_loc)
    storage_loc.mkdir(exist_ok=True)

    # check if an existing .pkl file exists with the same name, otherwise load it
    if (storage_loc / (word2vec_mname + ".pkl")).exists():
        if verbose:
            print("Loading existing word2vec model from {}".format(word2vec_mname))
        model = pickle.load(open(storage_loc / (word2vec_mname + ".pkl"), "rb"))
    else:
        print("Download word2vec model - Google News corpus via gensim")
        model = api.load(word2vec_mname)
        pickle.dump(model, open(storage_loc / (word2vec_mname + ".pkl"), "wb"))
    return model


def estimate_TOC_pages(
    n_entries: int, title_thresh: int = 39, std_thresh: int = 60, verbose=False
):
    """
    estimate_TOC_pages - a naive estimator for the number of pages required for
                        generated PDF report. Probably should be replaced by a
                        function that computes the actual distance / space with
                        the font size etc.
    Parameters
    ----------
    n_entries : int, the number of entries in the report
    title_thresh : int, the number of entries that will be printed on a single
                    page
    std_thresh : int, the number of entries that will be printed on a single
                    page
    verbose : bool, default False, print the number of pages estimated

    Returns
    ------
    int, the number of pages to save for TOC
    """
    assert n_entries > 0, f"require >=1 entry to estimate # pages, got {n_entries}"
    title_num = int(math.ceil(n_entries / title_thresh))
    if title_num < 2:
        if verbose:
            print(f"\n# entries is {n_entries}, < title thresh {title_thresh}")
            print("will use one page for TOC\n")
        return 1
    else:
        std_pg_entries = n_entries - title_thresh  # get entries not on the title page
        std_num = int(math.ceil(std_pg_entries / std_thresh))
        # TODO: find out what std_thresh actually is moving from 2->3 pages
        # and update
        toc_pgs = 1 + std_num
        if verbose:
            print(f"\n# entries is {n_entries}, > title thresh {title_thresh}")
            print(f"will use {toc_pgs} pages for TOC\n")
        return 1 + std_num


def load_punkt(nltk_usepunkt=True):
    """
    load_punkt - loads the punkt sentence tokenizer from nltk.

    Parameters
    ----------
    nltk_usepunkt : bool, optional, default True

    Returns
    -------
    punkt_sent_tokenizer : PunktSentenceTokenizer   # nltk.tokenize.punkt.PunktSentenceTokenizer
    """
    try:
        nltk.data.find("tokenizers/punkt")
        sent_detector = (
            nltk.data.load("tokenizers/punkt/english.pickle") if nltk_usepunkt else None
        )
    except LookupError:
        # if punkt is not installed, download it
        nltk.download("punkt")
        sent_detector = (
            nltk.data.load("tokenizers/punkt/english.pickle") if nltk_usepunkt else None
        )

    return sent_detector


def split_to_pars(
    full_text,
    wordvectors,
    use_punkt=False,
    sent_detector=None,
    segment_len=5,
    verbose=False,
):
    """
    split_to_pars - splits a string into paragraphs using the punkt sentence tokenizer or default

    Parameters
    ----------
    full_text : str, the full text to split into paragraphs
    wordvectors : Word2Vec model, the word2vec model to use for word embeddings
    use_punkt : bool, optional, default False
    sent_detector : [type], optional, default None, the punkt sentence tokenizer to use for splitting
    segment_len : int, optional, default 5, the length of the segments to use for word embeddings
    verbose : bool, optional, default False

    Returns
    -------
    list of str, the paragraphs of the text
    """

    sent_detector = (
        load_punkt(use_punkt) if use_punkt and sent_detector is None else sent_detector
    )
    if use_punkt:
        sentenced_text = sent_detector.tokenize(full_text.strip())
    else:
        sentence_tokenizer = SimpleSentenceTokenizer()
        sentenced_text = sentence_tokenizer(full_text)

    if isinstance(sentenced_text, str):
        sentenced_text = sentenced_text.split(".")

    # both parameters definitely will change depending on application
    if len(sentenced_text) > segment_len:
        vecr = CountVectorizer(vocabulary=wordvectors.index)
        sentence_vectors = vecr.transform(sentenced_text).dot(wordvectors)
        try:
            penalty = get_penalty([sentence_vectors], segment_len)
            optimal_segmentation = split_optimal(sentence_vectors, penalty)
            segmented_text = get_segments(sentenced_text, optimal_segmentation)
            paragraphs = [" ".join(segment) for segment in segmented_text]

            # segmented_text is a list of lists. The first level is segment and the second is sentences in the segment
        except Exception as e:
            warnings.warn(f"Error in split_to_pars: {e}, returning full text")
            segmented_text = sentenced_text
            paragraphs = segmented_text

    else:
        # the input text has less sentences than desired segments.
        # Just return the sentenced text
        warnings.warn(
            f"Input text has less sentences than desired segments. Returning full text"
        )
        paragraphs = sentenced_text
    # strip whitespace from the beginning and end of each paragraph
    paragraphs = [p.strip(" ") for p in paragraphs]
    return paragraphs


def print_pdf_margin_info(the_pdf):
    """
    print_pdf_margin_info - prints the margins of the pdf

    Parameters
    ----------
    the_pdf : FPDF object, the pdf to print margins for
    """
    pdf_height = the_pdf.h
    pdf_width = the_pdf.w
    left_margin = the_pdf.l_margin
    right_margin = the_pdf.r_margin
    writeable_space = math.floor(pdf_width - left_margin - right_margin)
    print(
        f"PDF dimensions are {round(pdf_width,3)} mm wide by {round(pdf_height,3)} mm tall"
    )
    print(
        f"width: {round(pdf_width,2)} mm ",
        f"The margins are  R: {right_margin} mm,  L: {left_margin} mm ",
    )
    print(f"the total useable space is {writeable_space} mm ")


def p_outside(pdf, text, **kwargs):
    """
    p_outside - prints text outside the margins of the pdf

    """
    # if you want to set align="C" you need to pass it in as a **kwargs
    pdf.multi_cell(w=pdf.epw, h=pdf.font_size, txt=text, ln=1, **kwargs)


def render_toc(pdf, outline, title="Table of Contents", verbose=False):
    """
    render_toc - renders the outline of the pdf

    Parameters
    ----------
    pdf : FPDF object, the pdf to render the outline for
    outline : list of dicts, the outline to render
    """
    # configure params
    pdf.y += 20
    pdf.set_x(pdf.l_margin)  # was in some weird location
    pdf.set_font("Helvetica", size=16)
    pdf.set_text_color(0)
    pdf.underline = False
    p_outside(pdf, title, align="L")  # write title of TOC
    pdf.y += 15
    pdf.set_font("Courier", size=12)
    # iterate through items
    for section in tqdm(outline):
        link = pdf.add_link()
        pdf.set_link(link, page=section.page_number)
        # get rid of spaces and the word chapter, keep only the first 20 chars

        upd_section_name = section.name
        upd_section_name = upd_section_name.replace("Chapter", "")
        if len(upd_section_name) > 45:
            upd_section_name = upd_section_name[:45]

        text = f'{" " * section.level * 2} {upd_section_name}'
        text += f' {"." * (60 - section.level * 2 - len(upd_section_name))} {section.page_number}'
        pdf.multi_cell(w=pdf.epw, h=pdf.font_size, txt=text, ln=1, align="C", link=link)
